# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Utitilies module."""
from __future__ import annotations

import collections
from enum import Enum
import ipaddress
import os
from pathlib import Path
import re
import socket
import sys
from typing import (
    DefaultDict,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    OrderedDict,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from ansys.optislang.core import FIRST_SUPPORTED_VERSION

VersionMapping = Dict[int, Path]


T = TypeVar("T", bound=Enum)


def enum_from_str(
    string: str,
    enum_class: Type[T],
    replace: Optional[Tuple[str, str]] = None,
    upper_case: bool = True,
) -> T:
    """Convert string to enumeration.

    Parameters
    ----------
    string: str
        String to be converted.
    enum_class: Enum
        Enumeration type, upper case enumeration items are expected.
    replace: Optional[Tuple[str, str]], optional
        Characters to be replaced in given ``string``.
            Tuple[0]: Replace from.
            Tuple[1]: Replace to.
    upper_case: bool, optional
        Whether string should be converted to upper_case. Defaults to `True`.

    Returns
    -------
    Enum
        Instance of given enumeration type.

    Raises
    ------
    TypeError
        Raised when the type of the ``string`` or the enum_class is invalid.
    ValueError
        Raised when the value for the ``string`` is invalid.
    """
    if not isinstance(string, str):
        raise TypeError(f"String was expected, but `{type(string)}` was given.")
    if not issubclass(enum_class, Enum):
        raise TypeError(f"Enumeration class was expected, but `{type(enum_class)}` was given.")
    if replace is not None:
        string = string.replace(replace[0], replace[1])
    if upper_case:
        string = string.upper()
    try:
        return enum_class[string]
    except:
        raise ValueError(f"{string} is not a member of {enum_class.__name__}.")


def get_osl_exec(osl_version: Optional[Union[int, str]] = None) -> Optional[Tuple[int, Path]]:
    """Get the path to the optiSLang executable file.

    Parameters
    ----------
    osl_version : Optional[Union[int, str]], optional
        optiSLang version in a three-digit format like this ``221``. The default
        is ``None``, in which case the latest installed version is used.

    Returns
    -------
    Optional[Tuple[int, pathlib.Path]]
        optiSLang version and path to the corresponding executable file, if it exists.
        If both Ansys and standalone installations are present, the latest Ansys
        installation is returned. If no executable file is found for a specified
        version, ``None`` is returned.

    Raises
    ------
    NotImplementedError
        Raised when the operating system is not supported.
    """
    osl_execs = find_all_osl_exec()
    if len(osl_execs) == 0:
        return None

    if isinstance(osl_version, str):
        try:
            osl_version = int(osl_version)
        except ValueError:
            return None

    if osl_version is not None:
        return (osl_version, osl_execs[osl_version][0]) if osl_version in osl_execs else None
    else:
        # It is assumed that the first item corresponds to the latest optiSLang version
        osl_version = next(iter(osl_execs.keys()))
        return (osl_version, osl_execs[osl_version][0])


def find_all_osl_exec() -> OrderedDict[int, Tuple[Path, ...]]:
    """Find all optiSLang executable files.

    Returns
    -------
    OrderedDict[int, Tuple[Path, ...]]
        Ordered dictionary of found optiSLang executables. The dictionary key corresponds
        to the optiSLang version and value to the tuple of the optiSLang executable paths.
        Ansys installations are first in the tuple, followed by standalone installations.
        The dictionary is sorted by the version number in descending order.

    Raises
    ------
    NotImplementedError
        Raised when the operating system is not supported.
    """
    # windows
    if os.name == "nt":
        return _find_all_osl_exec_in_windows()
    # linux
    elif os.name == "posix":
        return _find_all_osl_exec_in_posix()
    # another os
    else:
        raise NotImplementedError(f"Unsupported OS {os.name}.")


def get_osl_opx_import_script(osl_executable: Optional[Union[str, Path]] = None) -> Optional[Path]:
    """Get the path to the optiSLang OPX import script file.

    Parameters
    ----------
    osl_executable : Optional[Union[str, pathlib.Path]], optional
        optiSLang executable path to use as a reference for locating the import script file.
        The default is ``None``, in which case the default optiSLang executable file is used.

    Returns
    -------
    Optional[Path]
        Path to the optiSLang OPX import script file, if location succeeded,
        ``None`` is returned otherwise.

    Raises
    ------
    NotImplementedError
        Raised when the operating system is not supported.
    """
    if osl_executable is None:
        installed_version = get_osl_exec()
        if installed_version is not None:
            osl_executable = installed_version[1]

    if osl_executable is not None:
        osl_opx_import_script_path = (
            Path(osl_executable).parent / "tools" / "import" / "opx" / "convert_opx_to_opf.py"
        )
        if osl_opx_import_script_path.is_file():
            return osl_opx_import_script_path

    return None


def _find_all_osl_exec_in_windows() -> OrderedDict[int, Tuple[Path, ...]]:
    """Find all optiSLang executable files on Windows.

    Returns
    -------
    OrderedDict[int, Tuple[Path, ...]]
        Ordered dictionary of found optiSLang executables. The dictionary key corresponds
        to the optiSLang version and value to the tuple of the optiSLang executable paths.
        The dictionary is sorted by the version number in descending order.
    """
    all_osl_execs = _merge_osl_exec_dicts(
        [
            _find_ansys_osl_execs_in_envars("optislang.com"),
            _find_ansys_osl_execs_in_windows_program_files(),
            _find_standalone_osl_execs_in_windows(),
        ]
    )

    return _sort_osl_execs(all_osl_execs)


def _find_all_osl_exec_in_posix() -> OrderedDict[int, Tuple[Path, ...]]:
    """Find all optiSLang executable files on POSIX-compliant systems.

    Returns
    -------
    OrderedDict[int, Tuple[pathlib.Path, ...]]
        Ordered dictionary of found optiSLang executables. The dictionary key corresponds
        to the optiSLang version and value to the tuple of the optiSLang executable paths.
        The dictionary is sorted by the version number in descending order.
    """
    all_osl_execs = _merge_osl_exec_dicts(
        [
            _find_ansys_osl_execs_in_envars("optislang"),
            _find_ansys_osl_execs_in_posix(),
            _find_standalone_osl_execs_in_posix(),
        ]
    )

    return _sort_osl_execs(all_osl_execs)


def _find_ansys_osl_execs_in_envars(exec_name: str) -> VersionMapping:
    """Find optiSLang executable files based on environmental variables.

    The Ansys ``AWP_ROOT`` environment variable is used to determine the root directory of the Ansys
    installation.

    Parameters
    ----------
    exec_name : str
        optiSLang executable name.

    Returns
    -------
    Dict[int, pathlib.Path]
        Dictionary of found optiSLang executables. The dictionary key is an optiSLang version.
        The dictionary value is a path to the corresponding optiSLang executable file.
    """
    osl_execs = {}
    for version, awp_root_value in iter_awp_roots():
        osl_exec_path = awp_root_value / "optiSLang" / exec_name
        if osl_exec_path.is_file() and version >= FIRST_SUPPORTED_VERSION:
            osl_execs[version] = osl_exec_path
    return osl_execs


def _find_ansys_osl_execs_in_windows_program_files() -> VersionMapping:
    """Find optiSLang executable files in the ``Program Files`` directory on Windows.

    This search is performed in the standard installation directory of Ansys products.

    Returns
    -------
    Dict[int, pathlib.Path]
        Dictionary of found optiSLang executables. The dictionary key is an optiSLang version.
        The dictionary value is a path to the corresponding optiSLang executable file.
    """
    osl_execs = {}
    program_files_path = _get_program_files_path()
    for ansys_version_dir in (program_files_path / "ANSYS Inc").glob("v*/"):
        try:
            ansys_version = int(ansys_version_dir.name[1:])
        except ValueError:
            continue
        if ansys_version >= FIRST_SUPPORTED_VERSION:
            osl_exec_path = ansys_version_dir / "optiSLang" / "optislang.com"
            if osl_exec_path.is_file():
                osl_execs[ansys_version] = osl_exec_path
    return osl_execs


def _find_standalone_osl_execs_in_windows() -> VersionMapping:
    """Find executable files of standalone optiSLang installations on Windows.

    Returns
    -------
    Dict[int, pathlib.Path]
        Dictionary of found optiSLang executables. The dctionary key is an optiSLang version.
        The dictionary value is a path to the corresponding optiSLang executable file.
    """
    osl_execs = {}

    root_install_dirs = [
        _get_program_files_path() / "Dynardo" / "Ansys optiSLang",
        _get_program_files_path() / "Ansys optiSLang",
    ]
    for root_install_dir in root_install_dirs:
        if root_install_dir.is_dir():
            for osl_version_dir in root_install_dir.glob("*/"):
                osl_exec_path = osl_version_dir / "optislang.com"
                if os.path.isfile(osl_exec_path):
                    # convert version fmt "yyyy Rx" to "yyx"
                    # TODO: add '$' at the end of regex pattern to disable dev version
                    match = re.findall("[2][0][1-9][0-9] R[1-9]", str(osl_version_dir))
                    if match:
                        try:
                            ansys_version = int(match[0][2:4] + match[0][6])
                        except ValueError:
                            continue
                        if ansys_version >= FIRST_SUPPORTED_VERSION:
                            osl_execs[ansys_version] = osl_exec_path
    return osl_execs


def _find_ansys_osl_execs_in_posix() -> VersionMapping:
    """Find optiSLang executable files in default Ansys paths on POSIX-compliant systems.

    This search is performed in standard installation paths of Ansys products.

    Returns
    -------
    Dict[int, pathlib.Path]
        Dictionary of found optiSLang executables. The dictionary key is an optiSLang version.
        The dictionary value is a path to the corresponding optiSLang executable file.
    """
    osl_execs = {}
    base_path = None
    for ansys_dir in [Path("/usr/ansys_inc"), Path("/ansys_inc")]:
        if ansys_dir.is_dir():
            base_path = ansys_dir
    if base_path is not None:
        for ansys_version_dir in base_path.glob("v*/"):
            try:
                ansys_version = int(ansys_version_dir.name[1:])
            except ValueError:
                continue
            if ansys_version >= FIRST_SUPPORTED_VERSION:
                osl_exec_path = ansys_version_dir / "optiSLang" / "optislang"
                if osl_exec_path.is_file():
                    osl_execs[ansys_version] = osl_exec_path
    return osl_execs


def _find_standalone_osl_execs_in_posix() -> VersionMapping:
    """Find the executable files of standalone optiSLang installations on POSIX-compliant systems.

    Returns
    -------
    Dict[int, pathlib.Path]
        Dictionary of found optiSLang executables. The dictionary key is an optiSLang version.
        The dictionary value is a path to the corresponding optiSLang executable file.
    """
    osl_execs = {}
    root_install_dir = Path("/opt/dynardo")
    if root_install_dir.is_dir():
        for osl_version_dir in root_install_dir.glob("*/"):
            osl_exec_path = osl_version_dir / "optislang"
            if osl_exec_path.is_file():
                # convert version fmt "yyyy Rx" to "yyx"
                # TODO: add '$' at the end of regex pattern to disable dev version
                match = re.findall("[2][0][1-9][0-9]R[1-9]", str(osl_version_dir))
                if match:
                    try:
                        ansys_version = int(match[0][2:4] + match[0][5])
                    except ValueError:
                        continue
                    if ansys_version >= FIRST_SUPPORTED_VERSION:
                        osl_execs[ansys_version] = osl_exec_path
    return osl_execs


def _merge_osl_exec_dicts(osl_execs_dicts: Iterable[VersionMapping]) -> Dict[int, Tuple[Path, ...]]:
    """Merge dictionaries of optiSLang executable files into one dictionary.

    Parameters
    ----------
    osl_execs_dicts : Iterable[Dict[int, pathlib.Path]]
        Iterable of dictionaries of optiSLang executable files. In each dictionary,
        the key is an optiSLang version. The value is a corresponding path to the
        optiSLang executable file.

    Returns
    -------
    Dict[int, Tuple[pathlib.Path, ...]]
        Merged dictionary in which the key is an optiSLang version and the value is
        a tuple of paths to the corresponding optiSLang executable files.
    """
    osl_execs_merged: DefaultDict[int, List[Path]] = collections.defaultdict(list)

    for osl_execs in osl_execs_dicts:
        for osl_version, exec_path in osl_execs.items():
            if exec_path not in osl_execs_merged[osl_version]:
                osl_execs_merged[osl_version].append(exec_path)

    return {version: tuple(execs_paths) for version, execs_paths in osl_execs_merged.items()}


def _sort_osl_execs(osl_execs: Dict[int, Tuple[Path, ...]]) -> OrderedDict[int, Tuple[Path, ...]]:
    """Sort the dictionary of optiSLang executable files according to version in descending order.

    Parameters
    ----------
    osl_execs : Dict[int, Tuple[str, ...]]
        Dictionary of optiSLang executables. The dictionary key is an optiSLang version. The
        value is a tuple of paths to the optiSLang executable files.

    Returns
    -------
    OrderedDict[int, Tuple[str, ...]]
        Sorted ordered dictionary of optiSLang executable files. The dictionary key is an optiSLang
        version. The value is a tuple of paths to the optiSLang executable files.
    """
    return collections.OrderedDict(sorted(osl_execs.items(), reverse=True))


def _get_program_files_path() -> Path:
    """Get the path for the ``Program Files`` directory on Windows.

    Returns
    -------
    str
        Path to the ``Program Files`` directory.
    """
    return Path(os.environ.get("ProgramFiles", "C:\\Program Files"))


def iter_awp_roots() -> Iterator[Tuple[int, Path]]:
    """Iterate AWP_ROOTXXX environment variables.

    Yields
    -------
    tuple
        Ansys version and the respective root directory.
    """
    for varname, value in os.environ.copy().items():
        varname_match = re.fullmatch(r"AWP_ROOT([0-9]{3})", varname)
        if varname_match:
            yield int(varname_match.group(1)), Path(value)


def is_iron_python():
    """Whether current platform is IronPython."""
    return sys.platform == "cli"


def get_localhost_addresses() -> List[str]:
    """Get addresses of the localhost machine.

    Returns
    -------
    List[str]
        List of addresses of the localhost machine.
    """
    return [i[4][0] for i in socket.getaddrinfo(socket.gethostname(), None)]


def is_localhost(host: str) -> bool:
    """Determine whether the host is localhost.

    Parameters
    ----------
    host : str
        IPv4/v6 address or domain name of the host.

    Returns
    -------
    bool
        ``True`` if the provided ``host`` is localhost; ``False`` otherwise.
    """
    try:
        addrinfo = socket.getaddrinfo(host=host, port=None)
    except socket.gaierror:
        return False

    localhost_addresses = get_localhost_addresses()
    for _, _, _, _, sockaddr in addrinfo:
        try:
            ip = sockaddr[0]
            if not ipaddress.ip_address(ip).is_loopback and ip not in localhost_addresses:
                return False
        except ValueError:
            return False
    return True



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Create workflow."""
###############################################################################
# (c) 2021 ANSYS, Inc. Unauthorized use, distribution, or duplication is prohibited.

# create direct optimization using evolutionary algorithm flow for oscillator python example

# in batch:
#      optislang --new oscillator_py.opf --python oscillator_optimization_ea.py --run --batch

# in optiSlang:
#      open optiSLang and use 'run script file...' in optiSlang python console

# find a description of the oscillator example in optiSLang tutorial section
###############################################################################

###############################################################################
# Import needed modules
###############################################################################
import os

import PyNOA
from py_os_criterion import DesignStatus, PyOSCriterion, PyOSCriterionContainer

###############################################################################
# Create EA and PostProcessing
###############################################################################
EA = actors.EAActor("Evolutionary Algorithm")
EA.auto_save_mode = AS_ACTOR_FINISHED
add_actor(EA)

post_proc = actors.PostprocessingActor("PostProcessing")
add_actor(post_proc)

connect(EA, "OMDBPath", post_proc, "IMDBPath")

###############################################################################
# Create oscillator example @ example_system
###############################################################################
# get example path "OSL_EXAMPLES": .\pyoptislang\src\ansys\optislang\core\examples
example_system = EA
path = os.path.join(
    os.environ["OSL_EXAMPLES"], "00_run_script", "oscillator", "scripts", "_create_oscillator.py"
)
exec(open(path).read())
EA = example_system

###############################################################################
# Setup EA
###############################################################################
settings = EA.optimizer_settings
settings.set_start_pop_size(20)
settings.set_number_of_parents(10)
EA.optimizer_settings = settings

###############################################################################
# Setup Criteria
###############################################################################
criteria_exprs = PyOSCriterionContainer()
criteria_exprs.add("Objective", PyOSCriterion(DesignStatus.MIN, "x_max"))
criteria_exprs.add("Constraint", PyOSCriterion(DesignStatus.LESSEQUAL, "omega_damped", "8.0"))
EA.criteria = criteria_exprs

###############################################################################
# Setup Postprocessing
###############################################################################
post_proc.pp_mode = actors.PP3_AUTOMATIC
post_proc.show_pp_during_run = False
post_proc.show_reduced_pp_when_avaiable = False
post_proc.wait_for_pp_to_finish = True
post_proc.force_classic = False



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""This module contains available node types."""

from __future__ import annotations

from enum import Enum
from typing import Optional

from ansys.optislang.core.nodes import NodeClassType
from ansys.optislang.core.utils import enum_from_str


class AddinType(Enum):
    """Provides ``AddinType`` options."""

    BUILT_IN = 0
    ALGORITHM_PLUGIN = 1
    INTEGRATION_PLUGIN = 2
    PYTHON_BASED_ALGORITHM_PLUGIN = 3
    PYTHON_BASED_INTEGRATION_PLUGIN = 4
    PYTHON_BASED_MOP_NODE_PLUGIN = 5
    PYTHON_BASED_NODE_PLUGIN = 6

    @classmethod
    def from_str(cls, string: str) -> AddinType:
        """Convert string to an instance of the ``AddinType`` class.

        Parameters
        ----------
        string: str
            String to be converted.

        Returns
        -------
        AddinType
            Instance of the ``AddinType`` class.

        Raises
        ------
        TypeError
            Raised when an invalid type of ``string`` is given.
        ValueError
            Raised when an invalid value of ``string`` is given.
        """
        return enum_from_str(string=string, enum_class=cls, replace=(" ", "_"))


class NodeType:
    """Class containing information about node type."""

    def __init__(self, id: str, subtype: AddinType, osl_class_type: Optional[NodeClassType] = None):
        """Create an instance of the ``NodeType`` class.

        Parameters
        ----------
        id : str
            Type of node.
        subtype : AddinType
            Subtype of node.
        osl_class_type: Optional[NodeClassType], optional
            Type of class that should be created, by default ``None``.
        """
        self.__id = id
        self.__subtype = subtype
        self.__osl_class_type = osl_class_type

    def __str__(self):
        """Return formatted string."""
        return f"type: {self.id}, subtype: {self.subtype}"

    def __eq__(self, other: object) -> bool:
        """Object comparison."""
        if not isinstance(other, NodeType):
            return NotImplemented
        return self.id == other.id and self.subtype == other.subtype

    @property
    def id(self) -> str:
        """Type of node.

        Returns
        -------
        str
            Type of node.
        """
        return self.__id

    @property
    def osl_class_type(self) -> Optional[NodeClassType]:
        """Type of node class.

        Returns
        -------
        Optional[NodeClassType]
            Instance of the ``NodeClassType`` class, ``None`` if not specified.
        """
        return self.__osl_class_type

    @property
    def subtype(self) -> AddinType:
        """Subtype of node.

        Returns
        -------
        AddinType
            Instance of the ``AddinType`` class.
        """
        return self.__subtype


# region NODES
# region Builtins
AbaqusProcess = NodeType(
    id="AbaqusProcess", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
AmesimInput = NodeType(
    id="AmesimInput", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.NODE
)
AnsysAPDLParameterize = NodeType(
    id="AnsysAPDLParameterize",
    subtype=AddinType.BUILT_IN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
AnsysWorkbench = NodeType(
    id="AnsysWorkbench", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
AppendDesignsToBinFile = NodeType(
    id="AppendDesignsToBinFile", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.NODE
)
ASCMOsolver = NodeType(id="ASCMOsolver", subtype=AddinType.BUILT_IN)
BashScript = NodeType(
    id="BashScript", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
BatchScript = NodeType(
    id="BatchScript", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
Calculator = NodeType(id="Calculator", subtype=AddinType.BUILT_IN)
CalculatorSet = NodeType(
    id="CalculatorSet", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
CatiaParameterize = NodeType(
    id="CatiaParameterize",
    subtype=AddinType.BUILT_IN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
CatiaProcess = NodeType(
    id="CatiaProcess", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
CetolInput = NodeType(id="CetolInput", subtype=AddinType.BUILT_IN)
CFturboInput = NodeType(id="CFturboInput", subtype=AddinType.BUILT_IN)
Compare = NodeType(id="Compare", subtype=AddinType.BUILT_IN)
Custom = NodeType(id="Custom", subtype=AddinType.BUILT_IN)
CustomETKIntegration = NodeType(
    id="CustomETKIntegration",
    subtype=AddinType.BUILT_IN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
CustomIntegration = NodeType(
    id="CustomIntegration",
    subtype=AddinType.BUILT_IN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
CustomMop = NodeType(id="CustomMop", subtype=AddinType.BUILT_IN)
DataExport = NodeType(id="DataExport", subtype=AddinType.BUILT_IN)
DataImport = NodeType(id="DataImport", subtype=AddinType.BUILT_IN)
DataMining = NodeType(
    id="DataMining", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
DesignExport = NodeType(id="DesignExport", subtype=AddinType.BUILT_IN)
DesignImport = NodeType(id="DesignImport", subtype=AddinType.BUILT_IN)
DPS = NodeType(id="DPS", subtype=AddinType.BUILT_IN)
ETKAbaqus = NodeType(
    id="ETKAbaqus", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
ETKAdams = NodeType(
    id="ETKAdams", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
ETKAMESim = NodeType(
    id="ETKAMESim", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
ETKAnsys = NodeType(
    id="ETKAnsys", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
ETKAsciiOutput = NodeType(
    id="ETKAsciiOutput", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
ETKCetol = NodeType(
    id="ETKCetol", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
ETKCFturbo = NodeType(
    id="ETKCFturbo", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
ETKComplete = NodeType(
    id="ETKComplete", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
ETKEdyson = NodeType(
    id="ETKEdyson", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
ETKExtOut = NodeType(
    id="ETKExtOut", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
ETKFloEFD = NodeType(
    id="ETKFloEFD", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
ETKLSDYNA = NodeType(
    id="ETKLSDYNA", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
ETKMadymo = NodeType(
    id="ETKMadymo", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
ETKMidas = NodeType(
    id="ETKMidas", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
ETKSimPack = NodeType(
    id="ETKSimPack", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
ETKTurboOpt = NodeType(
    id="ETKTurboOpt", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
Excel = NodeType(
    id="Excel", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
FloEFDInput = NodeType(
    id="FloEFDInput", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
IntegrationPlugin = NodeType(
    id="IntegrationPlugin",
    subtype=AddinType.BUILT_IN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
LSDynaParameterize = NodeType(
    id="LSDynaParameterize",
    subtype=AddinType.BUILT_IN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
Matlab = NodeType(
    id="Matlab", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
Mop = NodeType(id="Mop", subtype=AddinType.BUILT_IN)
Mopsolver = NodeType(
    id="Mopsolver", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
MultiplasParameterize = NodeType(id="MultiplasParameterize", subtype=AddinType.BUILT_IN)
Octave = NodeType(
    id="Octave", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
OOCalc = NodeType(id="OOCalc", subtype=AddinType.BUILT_IN)
Parameterize = NodeType(
    id="Parameterize", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
Path = NodeType(id="Path", subtype=AddinType.BUILT_IN)
PDM = NodeType(id="PDM", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE)
PDMReceive = NodeType(
    id="PDMReceive", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
PDMSend = NodeType(
    id="PDMSend", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
PerlScript = NodeType(
    id="PerlScript", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
PMop = NodeType(id="PMop", subtype=AddinType.BUILT_IN)
PMOPPostprocessing = NodeType(id="PMOPPostprocessing", subtype=AddinType.BUILT_IN)
PMopsolver = NodeType(id="PMopsolver", subtype=AddinType.BUILT_IN)
Postprocessing = NodeType(id="Postprocessing", subtype=AddinType.BUILT_IN)
Process = NodeType(id="Process", subtype=AddinType.BUILT_IN)
ProEParameterize = NodeType(
    id="ProEParameterize", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
ProEProcess = NodeType(
    id="ProEProcess", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
ProxySolver = NodeType(
    id="ProxySolver", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.PROXY_SOLVER
)
Python2 = NodeType(
    id="Python2", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
PythonScript = NodeType(
    id="PythonScript", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
SimPackInput = NodeType(
    id="SimPackInput", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
SimulationX = NodeType(id="SimulationX", subtype=AddinType.BUILT_IN)
SolverTemplate = NodeType(id="SolverTemplate", subtype=AddinType.BUILT_IN)
SoSGenerate = NodeType(id="SoSGenerate", subtype=AddinType.BUILT_IN)
SoSPostprocessing = NodeType(id="SoSPostprocessing", subtype=AddinType.BUILT_IN)
String = NodeType(id="String", subtype=AddinType.BUILT_IN)
TaggedParametersParameterize = NodeType(
    id="TaggedParametersParameterize",
    subtype=AddinType.BUILT_IN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
TurboOptInput = NodeType(
    id="TurboOptInput", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.INTEGRATION_NODE
)
Variable = NodeType(id="Variable", subtype=AddinType.BUILT_IN)
VariantMonitoring = NodeType(id="VariantMonitoring", subtype=AddinType.BUILT_IN)
VCollabProcess = NodeType(id="VCollabProcess", subtype=AddinType.BUILT_IN)
Wait = NodeType(id="Wait", subtype=AddinType.BUILT_IN)
# endregion
# region Integration plugins
awb2_plugin = NodeType(
    id="awb2_plugin",
    subtype=AddinType.INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
discovery_plugin = NodeType(
    id="discovery_plugin",
    subtype=AddinType.INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
lsdyna_plugin = NodeType(
    id="lsdyna_plugin",
    subtype=AddinType.INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
optislang_node = NodeType(
    id="optislang_node",
    subtype=AddinType.INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
spaceclaim_plugin = NodeType(
    id="spaceclaim_plugin",
    subtype=AddinType.INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
speos_plugin = NodeType(
    id="speos_plugin",
    subtype=AddinType.INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
# endregion
# region Python based integration plugins
AEDT2 = NodeType(
    id="AEDT2",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
AEDT2_lsdso = NodeType(
    id="AEDT2_lsdso",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
ANSA_input = NodeType(
    id="ANSA_input",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
ANSA_output = NodeType(
    id="ANSA_output",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
AmplitudesFromField_SoS = NodeType(
    id="AmplitudesFromField_SoS",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
AxSTREAM = NodeType(
    id="AxSTREAM",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
CAD_CATIA = NodeType(
    id="CAD_CATIA",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
CAD_Creo = NodeType(
    id="CAD_Creo",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
CAD_Inventor = NodeType(
    id="CAD_Inventor",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
CAD_NX = NodeType(
    id="CAD_NX",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
CAESES_input = NodeType(
    id="CAESES_input",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
CFX_Partitioner = NodeType(
    id="CFX-Partitioner",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
CFX_Partitioner_v3 = NodeType(
    id="CFX-Partitioner-v3",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
CFX_Pre = NodeType(
    id="CFX-Pre",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
CFX_Pre_v3 = NodeType(
    id="CFX-Pre-v3",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
CFX_Solver = NodeType(
    id="CFX-Solver",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
CFX_Solver_v3 = NodeType(
    id="CFX-Solver-v3",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
CFturbo_input = NodeType(
    id="CFturbo_input",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
COMSOL2 = NodeType(
    id="COMSOL2",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
COMSOL_input = NodeType(
    id="COMSOL_input",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
COMSOL_output = NodeType(
    id="COMSOL_output",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
Convert_OMDB_to_BIN_SoS = NodeType(
    id="Convert_OMDB_to_BIN_SoS",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
ETK_nD = NodeType(
    id="ETK_nD",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
FMU_SoS = NodeType(
    id="FMU_SoS",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
Field_Data_Collector = NodeType(
    id="Field_Data_Collector",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
Field_MOP_2D_Nested_DOE_SoS = NodeType(
    id="Field_MOP_2D_Nested_DOE_SoS",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
Field_MOP_ANSYSMECH_SoS = NodeType(
    id="Field_MOP_ANSYSMECH_SoS",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
FloEFD_input = NodeType(
    id="FloEFD_input",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
FloEFD_output = NodeType(
    id="FloEFD_output",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
Fluent = NodeType(
    id="Fluent",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
Fluent_mesher = NodeType(
    id="Fluent_mesher",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
Fluent_solver = NodeType(
    id="Fluent_solver",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
Flux_input = NodeType(
    id="Flux_input",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
GTSUITE_input = NodeType(
    id="GTSUITE_input",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
GTSUITE_output = NodeType(
    id="GTSUITE_output",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
Generate_SoS = NodeType(
    id="Generate_SoS",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
GeoDict_input = NodeType(
    id="GeoDict_input",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
GeoDict_output = NodeType(
    id="GeoDict_output",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
IPG_Automotive = NodeType(
    id="IPG_Automotive",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
JMAG_Designer_input = NodeType(
    id="JMAG_Designer_input",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
JMAG_Designer_output = NodeType(
    id="JMAG_Designer_output",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
JMAG_Designer_solve = NodeType(
    id="JMAG_Designer_solve",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
JSON_input = NodeType(
    id="JSON_input",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
JSON_output = NodeType(
    id="JSON_output",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
KULI = NodeType(
    id="KULI",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
Lumerical = NodeType(
    id="Lumerical",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
META_output = NodeType(
    id="META_output",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
Matlab_mat_input = NodeType(
    id="Matlab_mat_input",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
Matlab_mat_output = NodeType(
    id="Matlab_mat_output",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
ModelCenter = NodeType(
    id="ModelCenter",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
MotorCAD_input = NodeType(
    id="MotorCAD_input",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
MotorCAD_output = NodeType(
    id="MotorCAD_output",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
MotorCAD_solve = NodeType(
    id="MotorCAD_solve",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
NASTRAN = NodeType(
    id="NASTRAN",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
OpticStudio = NodeType(
    id="OpticStudio",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
PuTTY_SSH = NodeType(
    id="PuTTY_SSH",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
ROCKY_input = NodeType(
    id="ROCKY_input",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
ROCKY_output = NodeType(
    id="ROCKY_output",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
SPEOSCore = NodeType(
    id="SPEOSCore",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
SPEOS_Report_Reader = NodeType(
    id="SPEOS_Report_Reader",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
SimulationX_SXOA = NodeType(
    id="SimulationX_SXOA",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
Viewer_SoS = NodeType(
    id="Viewer_SoS",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
VirtualLab_input = NodeType(
    id="VirtualLab_input",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
VirtualLab_output = NodeType(
    id="VirtualLab_output",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
ZEMAX = NodeType(
    id="ZEMAX",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
ZEMAX_input = NodeType(
    id="ZEMAX_input",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
ZEMAX_output = NodeType(
    id="ZEMAX_output",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
ZEMAX_solve = NodeType(
    id="ZEMAX_solve",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
optislang_omdb = NodeType(
    id="optislang_omdb",
    subtype=AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
    osl_class_type=NodeClassType.INTEGRATION_NODE,
)
# endregion
# region Python based MOP node plugins
MOP = NodeType(id="MOP", subtype=AddinType.PYTHON_BASED_MOP_NODE_PLUGIN)
# endregion
# endregion

# region SYSTEMS
# region Builtins
While = NodeType(id="While", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.SYSTEM)
# endregion
# endregion

# region PARAMETRIC SYSTEMS
# region Builtins
AlgorithmSystemPlugin = NodeType(
    id="AlgorithmSystemPlugin",
    subtype=AddinType.BUILT_IN,
    osl_class_type=NodeClassType.PARAMETRIC_SYSTEM,
)
AMOP = NodeType(
    id="AMOP", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.PARAMETRIC_SYSTEM
)
ARSM = NodeType(
    id="ARSM", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.PARAMETRIC_SYSTEM
)
CustomAlgorithm = NodeType(
    id="CustomAlgorithm", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.PARAMETRIC_SYSTEM
)
EA = NodeType(id="EA", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.PARAMETRIC_SYSTEM)
Memetic = NodeType(
    id="Memetic", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.PARAMETRIC_SYSTEM
)
NLPQLP = NodeType(
    id="NLPQLP", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.PARAMETRIC_SYSTEM
)
NOA2 = NodeType(
    id="NOA2", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.PARAMETRIC_SYSTEM
)
ParametricSystem = NodeType(
    id="ParametricSystem",
    subtype=AddinType.BUILT_IN,
    osl_class_type=NodeClassType.PARAMETRIC_SYSTEM,
)
PSO = NodeType(id="PSO", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.PARAMETRIC_SYSTEM)
Reevaluate = NodeType(
    id="Reevaluate", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.PARAMETRIC_SYSTEM
)
ReliabilityARSM = NodeType(
    id="ReliabilityARSM", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.PARAMETRIC_SYSTEM
)
ReliabilityAS = NodeType(
    id="ReliabilityAS", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.PARAMETRIC_SYSTEM
)
ReliabilityDS = NodeType(
    id="ReliabilityDS", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.PARAMETRIC_SYSTEM
)
ReliabilityFORM = NodeType(
    id="ReliabilityFORM", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.PARAMETRIC_SYSTEM
)
ReliabilityISPUD = NodeType(
    id="ReliabilityISPUD",
    subtype=AddinType.BUILT_IN,
    osl_class_type=NodeClassType.PARAMETRIC_SYSTEM,
)
ReliabilityMC = NodeType(
    id="ReliabilityMC", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.PARAMETRIC_SYSTEM
)
Robustness = NodeType(
    id="Robustness", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.PARAMETRIC_SYSTEM
)
SDI = NodeType(id="SDI", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.PARAMETRIC_SYSTEM)
Sensitivity = NodeType(
    id="Sensitivity", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.PARAMETRIC_SYSTEM
)
SIMPLEX = NodeType(
    id="SIMPLEX", subtype=AddinType.BUILT_IN, osl_class_type=NodeClassType.PARAMETRIC_SYSTEM
)
# endregion
# region Python based algorithm plugins
BASS = NodeType(
    id="BASS",
    subtype=AddinType.PYTHON_BASED_ALGORITHM_PLUGIN,
    osl_class_type=NodeClassType.PARAMETRIC_SYSTEM,
)
DXAMO = NodeType(
    id="DXAMO",
    subtype=AddinType.PYTHON_BASED_ALGORITHM_PLUGIN,
    osl_class_type=NodeClassType.PARAMETRIC_SYSTEM,
)
DXASO = NodeType(
    id="DXASO",
    subtype=AddinType.PYTHON_BASED_ALGORITHM_PLUGIN,
    osl_class_type=NodeClassType.PARAMETRIC_SYSTEM,
)
DXMISQP = NodeType(
    id="DXMISQP",
    subtype=AddinType.PYTHON_BASED_ALGORITHM_PLUGIN,
    osl_class_type=NodeClassType.PARAMETRIC_SYSTEM,
)
DXUPEGO = NodeType(
    id="DXUPEGO",
    subtype=AddinType.PYTHON_BASED_ALGORITHM_PLUGIN,
    osl_class_type=NodeClassType.PARAMETRIC_SYSTEM,
)
GLAD = NodeType(
    id="GLAD",
    subtype=AddinType.PYTHON_BASED_ALGORITHM_PLUGIN,
    osl_class_type=NodeClassType.PARAMETRIC_SYSTEM,
)
OCO = NodeType(
    id="OCO",
    subtype=AddinType.PYTHON_BASED_ALGORITHM_PLUGIN,
    osl_class_type=NodeClassType.PARAMETRIC_SYSTEM,
)
PIBO = NodeType(
    id="PIBO",
    subtype=AddinType.PYTHON_BASED_ALGORITHM_PLUGIN,
    osl_class_type=NodeClassType.PARAMETRIC_SYSTEM,
)
Replace_constant_parameter = NodeType(
    id="Replace_constant_parameter",
    subtype=AddinType.PYTHON_BASED_ALGORITHM_PLUGIN,
    osl_class_type=NodeClassType.PARAMETRIC_SYSTEM,
)
Unigene_EA = NodeType(
    id="Unigene_EA",
    subtype=AddinType.PYTHON_BASED_ALGORITHM_PLUGIN,
    osl_class_type=NodeClassType.PARAMETRIC_SYSTEM,
)
while_loop = NodeType(
    id="while_loop",
    subtype=AddinType.PYTHON_BASED_ALGORITHM_PLUGIN,
    osl_class_type=NodeClassType.PARAMETRIC_SYSTEM,
)
# endregion
# endregion


def get_node_type_from_str(node_id: str) -> NodeType:
    """Create instance of ``NodeType`` from string.

    Parameters
    ----------
    node_id: str
        Actor type in optiSLang server output format.

    Returns
    -------
    NodeType
        Instance of ``NodeType`` class.
    """
    customs = {
        "AlgorithmSystem_": AddinType.PYTHON_BASED_ALGORITHM_PLUGIN,
        "AlgorithmSystemPlugin_": AddinType.ALGORITHM_PLUGIN,
        "Custom_": AddinType.PYTHON_BASED_NODE_PLUGIN,
        "CustomETKIntegration_": "",  # TODO: append addin type
        "CustomIntegration_": AddinType.PYTHON_BASED_INTEGRATION_PLUGIN,
        "CustomMop_": AddinType.PYTHON_BASED_MOP_NODE_PLUGIN,
        "IntegrationPlugin_": AddinType.INTEGRATION_PLUGIN,
    }
    module_constants = {
        key: value
        for key, value in globals().items()
        if not (callable(value) or isinstance(value, type) or key.startswith("__"))
    }
    module_constants.pop("annotations")

    if module_constants.get(node_id, None) is not None:
        return module_constants[node_id]
    elif node_id in customs.keys():
        id_ = node_id[:-1]
        subtype = AddinType.BUILT_IN
    else:
        was_found = False
        for custom in customs.keys():
            if node_id.startswith(custom):
                id_ = node_id.replace(custom, "")
                subtype = customs[custom]  # type: ignore
                was_found = True
                break
        if not was_found:
            id_ = node_id
            subtype = AddinType.BUILT_IN
    return NodeType(id=id_, subtype=subtype)



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Return free ports."""
# (c) 2021 ANSYS, Inc. Unauthorized use, distribution, or duplication is prohibited.
#
# Check for free ports in a specific range (default=49690-49700)

import os
import socket

portrange = range(49690, 49701)

SOCK = None
PORT = 0

for port in portrange:
    try:
        server_address = ("localhost", port)
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        if not os.name == "posix":
            # set SO_EXCLUSIVEADDRUSE  socket option to ensure that only this socket
            # will listen on the given port
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_EXCLUSIVEADDRUSE, 1)
        sock.bind(server_address)
        sock.listen(5)
        SOCK = sock
        PORT = port
        break
    except Exception as e:
        PORT = 0
        # print('Failed to listen on port {0}: {1}'.format(port, str(e)))

portrange_str = "({0}-{1})".format(portrange[0], portrange[len(portrange) - 1])
if PORT:
    print("Found free port {0} in range {1}".format(PORT, portrange_str))
else:
    print("No free port found in range {0}".format(portrange_str))

if SOCK is not None:
    try:
        SOCK.close()
    except Exception as ex:
        print("Exception closing socket: {0}".format(str(ex)))



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Create python2 node."""
###############################################################################
# (c) 2021 ANSYS, Inc. Unauthorized use, distribution, or duplication is prohibited.
#
# Create a Python2 node and add some source code. Add and remove parameters and
# responses. Add internal variables.
###############################################################################

from py_os_design import *

# Set up Parametric System and insert Python integration actor
param_system = actors.ParametricSystemActor("System")
add_actor(param_system)

pyactor = actors.Python2Actor("Python2")
param_system.add_actor(pyactor)

connect(param_system, "IODesign", pyactor, "IDesign")
connect(pyactor, "ODesign", param_system, "IIDesign")

# Set content of python actor
pyactor.source = "parameter_1 = 5.\nresponse_1 = parameter_1\npython_variable = 5.*parameter_1"

# Add parameters
pyactor.add_parameter("parameter_1", PyOSDesignEntry(5.0))
pyactor.add_parameter("temporary_parameter", PyOSDesignEntry(5.0))
# Remove parameters
pyactor.remove_parameter("temporary_parameter")

# Add responses
pyactor.add_response(
    ("response_2", PyOSDesignEntry(5.0)),
    actors.DerivedLocation("response_2", "parameter_1*response_1"),
)
pyactor.add_response(("response_1", PyOSDesignEntry(5.0)))
pyactor.add_response(("g", PyOSDesignEntry(5.0)))
# Remove responses
pyactor.remove_response("g")

# Add_expression/register_as_internal_location
# both lead to the same
pyactor.register_as_internal_location(
    "python_variable", "internal_from_py_source", PyOSDesignEntry(5.0)
)
pyactor.add_expression("c2", "internal_from_py_source_2", PyOSDesignEntry(5.0))
pyactor.register_as_internal_location(
    actors.DerivedLocation("d", "parameter_1 * response_1"), "d", PyOSDesignEntry(5.0)
)
pyactor.add_expression(
    actors.DerivedLocation("d2", "parameter_1 * response_1"), "d2", PyOSDesignEntry(5.0)
)
pyactor.register_as_internal_location(
    actors.DerivedLocation("f", "parameter_1 * response_1"), "f", PyOSDesignEntry(5.0)
)

# Remove expression/internal_location
pyactor.unregister_from_internal_locations("f")
pyactor.remove_expression("d2")

# Get parameters and print their names and values
params = PyOSDesignPoint()
pyactor.get_parameter(params)
print("Parameters:")
for name, value in params:
    print(name, value.get_scalar())

# Get responses and print their names and values
# Note that response values are updated after running the system
responses = PyOSDesignPoint()
pyactor.get_responses(responses)
print("Responses:")
for name, value in responses:
    print(name, value.get_scalar())



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Contains abstract ``Project`` class."""
from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Sequence, Tuple, Union

if TYPE_CHECKING:
    from ansys.optislang.core.io import RegisteredFile
    from ansys.optislang.core.nodes import RootSystem
    from ansys.optislang.core.project_parametric import Design


class Project(ABC):
    """Base class for classes which operate with active project."""

    @abstractmethod
    def __init__(self):  # pragma: no cover
        """``Project`` class is an abstract base class and cannot be instantiated."""
        pass

    @property
    @abstractmethod
    def root_system(self) -> RootSystem:  # pragma: no cover
        """Instance of the ``RootSystem`` class.

        Returns
        -------
        RootSystem
            Loaded project's root system.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        pass

    @property
    @abstractmethod
    def uid(self) -> str:  # pragma: no cover
        """Unique ID of the optiSLang project.

        Returns
        -------
        str
            Unique ID of the loaded project.
        """
        pass

    @abstractmethod
    def evaluate_design(
        self,
        design: Design,
    ) -> Design:  # pragma: no cover
        """Evaluate a design.

        Parameters
        ----------
        design: Design
            Instance of a ``Design`` class with defined parameters.

        Returns
        -------
        Design
            Evaluated design.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        pass

    @abstractmethod
    def get_description(self) -> Optional[str]:  # pragma: no cover
        """Get the description of the optiSLang project.

        Returns
        -------
        Optional[str]
            Description of the optiSLang project. If no project is loaded in optiSLang,
            ``None`` is returned.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        pass

    @abstractmethod
    def get_location(self) -> Optional[Path]:  # pragma: no cover
        """Get the path to the optiSLang project file.

        Returns
        -------
        Optional[pathlib.Path]
            Path to the optiSLang project file. If no project is loaded in the optiSLang,
            ``None`` is returned.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        pass

    @abstractmethod
    def get_name(self) -> str:  # pragma: no cover
        """Get the name of the optiSLang project.

        Returns
        -------
        str
            Name of the optiSLang project. If no project is loaded in the optiSLang,
            ``None`` is returned.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        pass

    @abstractmethod
    def get_reference_design(self) -> Design:  # pragma: no cover
        """Get a design with reference values of the parameters.

        Returns
        -------
        Design
            Instance of the ``Design`` class with defined parameters and reference values.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        pass

    @abstractmethod
    def get_registered_files(self) -> Tuple[RegisteredFile, ...]:  # pragma: no cover
        """Get all registered files in the current project.

        Returns
        -------
        Tuple[RegisteredFile, ...]
            Tuple with registered files.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        pass

    @abstractmethod
    def get_result_files(self) -> Tuple[RegisteredFile, ...]:  # pragma: no cover
        """Get result files.

        Returns
        -------
        Tuple[RegisteredFile, ...]
            Tuple with result files

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        pass

    @abstractmethod
    def get_status(self) -> str:  # pragma: no cover
        """Get the status of the optiSLang project.

        Returns
        -------
        str
            Status of the optiSLang project. If no project is loaded in optiSLang,
            ``None`` is returned.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        pass

    @abstractmethod
    def get_working_dir(self) -> Optional[Path]:  # pragma: no cover
        """Get the path to the optiSLang project's working directory.

        Returns
        -------
        Optional[pathlib.Path]
            Path to the optiSLang project's working directory. If no project is loaded
            in optiSLang, ``None`` is returned.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        pass

    @abstractmethod
    def reset(self) -> None:  # pragma: no cover
        """Reset the project.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        pass

    @abstractmethod
    def run_python_file(
        self,
        file_path: Union[str, Path],
        args: Union[Sequence[object], None] = None,
    ) -> Tuple[str, str]:  # pragma: no cover
        """Read a Python script from a file, load it in a project context, and run it.

        Parameters
        ----------
        file_path : Union[str, pathlib.Path]
            Path to the Python script file with the content to execute on the server.
        args : Sequence[object], None, optional
            Sequence of arguments to use in the Python script. The default is ``None``.

        Returns
        -------
        Tuple[str, str]
            STDOUT and STDERR from the executed Python script.

        Raises
        ------
        FileNotFoundError
            Raised when the specified Python script file does not exist.
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        pass

    @abstractmethod
    def run_python_script(
        self,
        script: str,
        args: Union[Sequence[object], None] = None,
    ) -> Tuple[str, str]:  # pragma: no cover
        """Load a Python script in a project context and run it.

        Parameters
        ----------
        script : str
            Python commands to execute on the server.
        args : Sequence[object], None, optional
            Sequence of arguments used in the Python script. The default
            is ``None``.

        Returns
        -------
        Tuple[str, str]
            STDOUT and STDERR from the executed Python script.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        pass

    @abstractmethod
    def start(
        self, wait_for_started: bool = True, wait_for_finished: bool = True
    ) -> None:  # pragma: no cover
        """Start project execution.

        Parameters
        ----------
        wait_for_started : bool, optional
            Determines whether this function call should wait on the optiSlang to start
            the command execution. I.e. don't continue on next line of python script
            after command was successfully sent to optiSLang but wait for execution of
            flow inside optiSLang to start.
            Defaults to ``True``.
        wait_for_finished : bool, optional
            Determines whether this function call should wait on the optiSlang to finish
            the command execution. I.e. don't continue on next line of python script
            after command was successfully sent to optiSLang but wait for execution of
            flow inside optiSLang to finish.
            This implicitly interprets wait_for_started as True.
            Defaults to ``True``.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with server.
        OslCommandError
            Raised when the command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        pass

    @abstractmethod
    def stop(self, wait_for_finished: bool = True) -> None:  # pragma: no cover
        """Stop project execution.

        Parameters
        ----------
        wait_for_finished : bool, optional
            Determines whether this function call should wait on the optiSlang to finish
            the command execution. I.e. don't continue on next line of python script after command
            was successfully sent to optiSLang but wait for execution of command inside optiSLang.
            Defaults to ``True``.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with server.
        OslCommandError
            Raised when the command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        pass

    # FUTURES:
    # @abstractmethod
    # TODO: Add this after it's fixed on optiSLang server side.
    # stop_gently method doesn't work properly in optiSLang 2023R1, therefore it was commented out
    # def stop_gently(self, wait_for_finished: bool = True) -> None:
    #     """Stop project execution after the current design is finished.

    #     Parameters
    #     ----------
    #     wait_for_finished : bool, optional
    #         Determines whether this function call should wait on the optiSlang to finish
    #         the command execution. I.e. don't continue on next line of python script after command
    #         was successfully sent to optiSLang but wait for execution of command inside optiSLang.
    #         Defaults to ``True``.

    #     Raises
    #     ------
    #     OslCommunicationError
    #         Raised when an error occurs while communicating with server.
    #     OslCommandError
    #         Raised when the command or query fails.
    #     TimeoutError
    #         Raised when the timeout float value expires.
    #     """



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
optiSLang.

core
"""

import importlib.metadata
import importlib.util
import os
import sys

from ansys.optislang.core.logging import OslLogger

LOG = OslLogger(loglevel="ERROR", log_to_file=False, log_to_stdout=True)
LOG.logger.debug("Loaded logging module as LOG")

__version__ = importlib.metadata.version(__name__.replace(".", "-"))

# First supported version of optiSLang: 2023R1
FIRST_SUPPORTED_VERSION = 231

from ansys.optislang.core.optislang import Optislang
from ansys.optislang.core.osl_process import OslServerProcess, ServerNotification

# Provide examples directory path
EXAMPLES_MODULE = "ansys.optislang.core.examples"
if spec := importlib.util.find_spec(EXAMPLES_MODULE):
    if spec.origin:
        os.environ["OSL_EXAMPLES"] = os.path.dirname(spec.origin)
    else:
        LOG.logger.warning(
            f"Could not set path to examples. Missing spec for module {EXAMPLES_MODULE}."
        )
else:
    LOG.logger.warning(
        f"Could not set path to examples. Missing origin for module {EXAMPLES_MODULE}."
    )



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Create workflow."""
###############################################################################
# (c) 2021 ANSYS, Inc. Unauthorized use, distribution, or duplication is prohibited.

# create parametric system for oscillator python example

# in batch:
#      optislang --new oscillator_py.opf --python oscillator_system_python.py --run --batch

# in optiSlang:
#      open optiSLang and use 'run script file...' in optiSlang python console

# find a description of the oscillator example in optiSLang tutorial section.
###############################################################################


###############################################################################
# Import needed modules
###############################################################################
import os

from dynardo_py_algorithms import ParameterType, RandomVariableType
from py_os_criterion import DesignStatus, PyOSCriterion, PyOSCriterionContainer
from py_os_design import PyOSDesignEntry, create_nvp_design_entry
from py_os_parameter import PyParameter
from stdcpp_python_export import double_list_to_vec

###############################################################################
# Create ParametricSystem and add python2-knot
###############################################################################
para_system = actors.ParametricSystemActor("oscillator")
add_actor(para_system)

py2 = actors.PythonActor("oscillator.py")
para_system.add_actor(py2)

connect(para_system, "IODesign", py2, "IDesign")
connect(py2, "ODesign", para_system, "IIDesign")

###############################################################################
# Setup python2-knot
###############################################################################
# get example path "OSL_EXAMPLES": .\pyoptislang\src\ansys\optislang\core\examples
path = os.path.join(
    os.environ["OSL_EXAMPLES"], "00_run_script", "oscillator", "files", "oscillator.py"
)
py2.path = Path(path)

py2.add_parameter("m", PyOSDesignEntry(1.0))
py2.add_parameter("k", PyOSDesignEntry(20.0))
py2.add_parameter("D", PyOSDesignEntry(0.02))
py2.add_parameter("Ekin", PyOSDesignEntry(10.0))

py2.add_response(create_nvp_design_entry("omega_damped", PyOSDesignEntry(4.47124)))
py2.add_response(create_nvp_design_entry("x_max", PyOSDesignEntry(0.62342)))

###############################################################################
# Setup ParameterManager
###############################################################################
params = para_system.parameter_manager

params.set_parameter_type("m", ParameterType.MIXED)
params.set_parameter_type("k", ParameterType.MIXED)
params.set_parameter_type("D", ParameterType.STOCHASTIC)
params.set_parameter_type("Ekin", ParameterType.STOCHASTIC)

params.set_bounds("m", PyOSDesignEntry(0.1), PyOSDesignEntry(5.0))
params.set_bounds("k", PyOSDesignEntry(10.0), PyOSDesignEntry(50.0))

params.set_ran_type("m", RandomVariableType.NORMAL)
params.set_ran_type("k", RandomVariableType.NORMAL)
params.set_ran_type("D", RandomVariableType.NORMAL)
params.set_ran_type("Ekin", RandomVariableType.NORMAL)

params.set_ran_parameters("m", double_list_to_vec([1.0, 0.02]))
params.set_ran_parameters("k", double_list_to_vec([20.0, 1.0]))
params.set_ran_parameters("D", double_list_to_vec([0.02, 0.002]))
params.set_ran_parameters("Ekin", double_list_to_vec([10.0, 1.0]))

para_system.parameter_manager = params

###############################################################################
# Setup Criteria
###############################################################################
criteria_exprs = PyOSCriterionContainer()
criteria_exprs.add("Objective", PyOSCriterion(DesignStatus.MIN, "x_max"))
criteria_exprs.add("Constraint", PyOSCriterion(DesignStatus.LESSEQUAL, "omega_damped", "8.0"))
para_system.criteria = criteria_exprs



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Contains abstract classes for obtaining and operating with project parametric."""
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, Any, Tuple

from ansys.optislang.core.project_parametric import Criterion, Parameter, Response

if TYPE_CHECKING:
    from ansys.optislang.core.osl_server import OslServer


class CriteriaManager:
    """Base classes for classes that obtains and operate with criteria."""

    @abstractmethod
    def __init__(self) -> None:  # pragma: no cover
        """``CriteriaManager`` class is an abstract base class and cannot be instantiated."""
        pass

    @abstractmethod
    def add_criterion(self, criterion: Criterion) -> None:  # pragma: no cover
        """Add criterion to the system.

        Returns
        -------
        Criterion
            Criterion to be created in the system.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        pass

    @abstractmethod
    def get_criteria(self) -> Tuple[Criterion, ...]:  # pragma: no cover
        """Get the criteria of the system.

        Returns
        -------
        Tuple[Criterion, ...]
            Tuple of the criterion for the system.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        pass

    @abstractmethod
    def get_criteria_names(self) -> Tuple[str, ...]:  # pragma: no cover
        """Get all criteria names.

        Returns
        -------
        Tuple[str, ...]
            Tuple of all criteria names.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        pass

    @abstractmethod
    def modify_criterion(self, criterion: Criterion) -> None:  # pragma: no cover
        """Modify criterion in the system.

        Parameters
        ----------
        criterion : Criterion
            Criterion to be modified. Criterion name is used as idefentifier.

        Raises
        ------
        NotImplementedError
            Raised when unsupported optiSLang server is used.
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        pass

    @abstractmethod
    def modify_criterion_property(
        self, criterion_name: str, property_name: str, property_value: Any
    ) -> None:  # pragma: no cover
        """Modify property of criterion in the system.

        Parameters
        ----------
        criterion_name : str
            Name of criterion to be modified.
        property_name: str
            Name of property to be modified.
        property_value: Any
            New value of the modified property.

        Raises
        ------
        NotImplementedError
            Raised when unsupported optiSLang server is used.
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        pass

    @abstractmethod
    def remove_all_criteria(self) -> None:  # pragma: no cover
        """Remove all criteria from the system.

        Raises
        ------
        NotImplementedError
            Raised when unsupported optiSLang server is used.
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        pass

    @abstractmethod
    def remove_criterion(self, criterion_name: str) -> None:  # pragma: no cover
        """Remove criterion from the system.

        Parameters
        ----------
        criterion_name : str
            Name of the criterion to be removed.

        Raises
        ------
        NotImplementedError
            Raised when unsupported optiSLang server is used.
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        pass


class ParameterManager:
    """Base classes for classes that obtains and operate with parameters."""

    @abstractmethod
    def __init__(self, uid: str, osl_server: OslServer) -> None:  # pragma: no cover
        """``ParameterManager`` class is an abstract base class and cannot be instantiated."""
        pass

    @abstractmethod
    def add_parameter(self, parameter: Parameter) -> None:  # pragma: no cover
        """Add parameter to the system.

        Returns
        -------
        Parameter
            Parameter to be created in the system.

        Raises
        ------
        NameError
            Raised when parameter with given name already exists.
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        pass

    @abstractmethod
    def get_parameters(self) -> Tuple[Parameter, ...]:  # pragma: no cover
        """Get the parameters of the system.

        Returns
        -------
        Tuple[Parameter, ...]
            Tuple of the parameters for the system.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        pass

    @abstractmethod
    def get_parameters_names(self) -> Tuple[str, ...]:  # pragma: no cover
        """Get all parameter names.

        Returns
        -------
        Tuple[str, ...]
            Tuple of all parameter names.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        pass

    @abstractmethod
    def modify_parameter(self, parameter: Parameter):  # pragma: no cover
        """Modify parameter in the system.

        Parameters
        ----------
        parameter: Parameter
            Parameter to be modified. Parameter name is used as identifier.

        Raises
        ------
        NameError
            Raised when the parameter with the given name doesn't exist.
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        pass

    @abstractmethod
    def modify_parameter_property(
        self, parameter_name: str, property_name: str, property_value: Any
    ) -> None:  # pragma: no cover
        """Modify property of parameter in the system.

        Parameters
        ----------
        parameter_name: str
            Name of the parameter to be modified.
        property_name: str
            Name of the property to be modified.
        property_value: Any
            New value of the modified property.

        Raises
        ------
        NameError
            Raised when the parameter with the given name doesn't exists.
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        pass

    @abstractmethod
    def remove_all_parameters(self) -> None:  # pragma: no cover
        """Remove all parameters from the system.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        pass

    @abstractmethod
    def remove_parameter(self, parameter_name: str) -> None:  # pragma: no cover
        """Remove parameter from the system.

        Parameters
        ----------
        parameter_name : str
            Name of the parameter to be removed.

        Raises
        ------
        NotImplementedError
            Raised when unsupported optiSLang server is used.
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        pass


class ResponseManager:
    """Base classes for classes that obtains and operate with responses."""

    @abstractmethod
    def __init__(self, uid: str, osl_server: OslServer) -> None:  # pragma: no cover
        """``ResponseManager`` class is an abstract base class and cannot be instantiated."""
        pass

    @abstractmethod
    def get_responses(self) -> Tuple[Response, ...]:  # pragma: no cover
        """Get the responses of the system.

        Returns
        -------
        Tuple[Criterion, ...]
            Tuple of the responses for the system.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        pass

    @abstractmethod
    def get_responses_names(self) -> Tuple[str, ...]:  # pragma: no cover
        """Get all responses names.

        Returns
        -------
        Tuple[str, ...]
            Tuple of all responses names.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        pass
